#!/usr/bin/perl
#
#  Copyright saturn@surrealchat.net
#  multiple feature-adds and code changes tabris@tabris.net
#
#  Licensed under the GNU Public License 
#  http://www.gnu.org/licenses/gpl.txt
#

package securitybot;

use strict;
no strict "refs";
use FreezeThaw;
use Time::Local;
#use Time::HiRes qw(sleep gettimeofday);
use DBI;

my $debug = 0;
my $master = "saturn";

#this stuff needs to be put into files
my $sbnick = "SecurityBot";
my $ident = 'Security';
my $gecos = 'Security Monitor (you are being monitored)';
my $umodes = '+BHSdopqz';
my $vhost = 'services.SC.bot';

my @except = ("localhost", "nascent.surrealchat.net", "cpe-65-30-240-15.mn.rr.com", "192.168.*.*"); #except doesn't work?
my @gexcept = ();

my %conf : shared;
my %sql_conf : shared;
my %torip : shared;

my %unwhois : shared;

my $dbh;

my (
	$add_spamfilter, $del_spamfilter, $add_tklban, $del_tklban,
	$del_expired_tklban, $get_expired_tklban,

	$get_tklban, $get_spamfilter,
	$get_all_tklban, $get_all_spamfilter,

	$check_opm,
);


sub begin {
	$dbh = DBI->connect(
		"DBI:mysql:".$sql_conf{'mysql-db'}.":mysql_server_prepare=1",
		$sql_conf{'mysql-user'},
		$sql_conf{'mysql-pass'},
		{
			AutoCommit => 1,
			RaiseError => 1,
			mysql_auto_reconnect => 1,
		}
	);
	# Prevent timeout
	$dbh->do("SET wait_timeout=(86400*365)");

	$add_tklban = $dbh->prepare_cached("REPLACE INTO tklban
		SET type=?, ident=?, host=?, setter=?, expire=?, time=?, reason=?");
	$del_tklban = $dbh->prepare_cached("DELETE FROM tklban WHERE type=? AND ident=? AND host=?");
	$add_spamfilter = $dbh->prepare_cached("REPLACE INTO spamfilter 
		SET target=?, action=?, setter=?, expire=?, time=?, bantime=?, reason=?, mask=?");
	$del_spamfilter = $dbh->prepare_cached("DELETE FROM spamfilter WHERE target=? AND action=? AND mask=?");

	$del_expired_tklban = $dbh->prepare_cached("DELETE FROM tklban WHERE expire <= UNIX_TIMESTAMP() AND expire!=0");
	$get_expired_tklban = $dbh->prepare_cached("SELECT type, ident, host, setter, expire, time, reason 
		FROM tklban WHERE expire <= UNIX_TIMESTAMP() AND expire!=0");

	$get_tklban = $dbh->prepare_cached("SELECT setter, expire, time, reason FROM tklban WHERE
		type=? AND ident=? AND host=?");
	$get_spamfilter = $dbh->prepare_cached("SELECT time, reason FROM spamfilter WHERE target=? AND action=? AND mask=?");

	$get_all_tklban = $dbh->prepare_cached("SELECT type, ident, host, setter, expire, time, reason
		FROM tklban ORDER BY type, time, host");
	$get_all_spamfilter = $dbh->prepare_cached("SELECT target, action, setter, expire, time, bantime, reason, mask, managed
		FROM spamfilter ORDER BY time, mask");

	$check_opm = $dbh->prepare_cached("SELECT 1 FROM opm WHERE ipaddr=?");
}

sub init {
	event::addhandler('SEOS', undef, undef, __PACKAGE__, "securitybot::register");
	event::addhandler('TKL', undef, undef, __PACKAGE__, "securitybot::handle_tkl");
	loadconf(0);
	$dbh = DBI->connect(
		"DBI:mysql:".$sql_conf{'mysql-db'}.":mysql_server_prepare=1",
		$sql_conf{'mysql-user'},
		$sql_conf{'mysql-pass'},
		{
			AutoCommit => 1,
			RaiseError => 1
		}
	);
	$dbh->do("TRUNCATE TABLE tklban");
	$dbh->do("TRUNCATE TABLE spamfilter");
	$dbh->disconnect();
}

=cut
my %snomasks = (
	e => 'Eyes Notice',
	v => 'VHost Notice',
	# They're prefixed already.
	#S => 'Spamfilter',
	o => 'Oper-up Notice',
);
=cut

sub snotice($$$) {
	my ($server, $type, $msg) = @_;
#	$type = $snomasks{$type};
#	net::privmsg($sbnick, $config::conf{diag}, ($type ? "[$type] " : '').$msg);
	net::privmsg($sbnick, $config::conf{diag}, $msg);
}

sub register {
	net::agent_connect($sbnick, $ident, $vhost, $umodes, $gecos);
	net::sqline($sbnick, 'Reserved for Services');
	event::addhandler('PRIVMSG', undef, $sbnick, __PACKAGE__, "securitybot::msghandle");
	event::addhandler('NOTICE', undef, $sbnick, __PACKAGE__, "securitybot::noticehandle");
	event::addhandler('SENDSNO', undef, undef, __PACKAGE__, "securitybot::snotice");

	
	if($conf{'EnableTor'} or $conf{'CTCPonConnect'} or $conf{'EnableOPM'}) {
		event::addhandler('NICKCONN', undef, undef, __PACKAGE__, 'securitybot::nickconn');
		event::addhandler('USERIP', undef, undef, __PACKAGE__, 'securitybot::userip');
	}
	
	net::join($sbnick, $config::conf{diag});
	net::setmode($sbnick, $config::conf{diag}, '+o', $sbnick);

	update_tor_list_timed(3540) if $conf{'EnableTor'};
	update_ss_timed(3300) if $conf{'EnableSS'};
	expire_tkl_timed();
};

sub nickconn {
	my ($rnick, $time, $ident, $host, $vhost, $server, $modes, $gecos, $ip) = @_[0,2..4,8,5,7,9,10];

	if($conf{'EnableTor'} or $conf{'EnableOPM'}) {
		if ($ip) {
			check_blacklists($rnick, $ip) or return;
		}
		else {
			$unwhois{lc $rnick} = 1;
			net::userip($rnick) unless module::is_loaded('services');
		}
	}

	if($conf{'CTCPonConnect'}) {
		my @ctcplist = split(/ /, $conf{'CTCPonConnect'});
		foreach my $ctcp_msg (@ctcplist) {
			if(uc($ctcp_msg) eq 'PING') {
				# This really should use Time::HiRes::gettimeofday
				# and the last param should be the usec portion
				net::ctcp($sbnick, $rnick, 'PING', time(), 0);
			} else {
				net::ctcp($sbnick, $rnick, uc($ctcp_msg));
			}
		}
	}
}

# XXX - we depend on nickserv to send userip to every user, so this won't work
# without services loaded
sub userip {
	my($src, $rnick, $ip) = @_;

	return unless($unwhois{lc $rnick});
	return unless($ip =~ /^\d{1,3}(\.\d{1,3}){3}$/);

	if($conf{'EnableTor'}) { check_tor($rnick, $ip) or return; }

	delete $unwhois{lc $rnick};
}

sub check_opm($) {
	my ($ip) = @_;
	$check_opm->execute($ip);
	my ($ret) = $check_opm->fetchrow_array();
	$check_opm->finish();
	return $ret;
}

sub mk_banreason($$) {
	my ($reason, $ip) = @_;
	$reason =~ s/\$/$ip/g;
	return $reason;
}

sub check_blacklists($$) {
	my ($rnick, $ip) = @_;
	
	if($conf{'EnableTor'} && $torip{$ip}) {
		net::zline($sbnick, $ip, $conf{'ProxyZlineTime'}, $conf{'TorZlineReason'});
		return 0;
	}

	if($conf{'EnableOPM'} && check_opm($ip)) {
		net::zline($sbnick, $ip, $conf{'ProxyZlineTime'}, mk_banreason($conf{'OPMZlineReason'}, $ip));
		return 0;
	}
	

	return 1;
}

sub update_tor_list_timed($) {
	my $time = shift;
	$time = 3600 unless $time;

	event::addtimer('', $time, __PACKAGE__, 'securitybot::update_tor_list_timed');
	
	update_tor_list();
}

sub update_tor_list() {
	net::privmsg($sbnick, $config::conf{diag}, " -- Loading Tor server list.");
	
	open TOR, '-|', ('wget -q -O - ' . $conf{'TorServer'});

	my %newtorip;
	while(my $l = <TOR>) {
		if($l =~ /^router (\S+) (\S+)/) {
			$newtorip{$2} = 1 unless $2 =~ /^(127|10|192\.168)\./;
		}
	}

	close TOR;

	my $torcount = scalar(keys(%newtorip));

	if($torcount > 0) {
		{
			lock %torip;
			%torip = %newtorip;
		}
		net::privmsg($sbnick, $config::conf{diag}, " -- Finished loading Tor server list - $torcount servers found.");
	} else {
		net::privmsg($sbnick, $config::conf{diag}, " -- Failed to load Tor server list, CHECK YOUR TorServer SETTING.");
	}
}

sub msghandle {
		my ($rnick, $dst, $msg) = @_;
		print join("\n", @_);
		my $user = { NICK => $rnick, AGENT => $sbnick };
		unless (adminserv::is_ircop($user)) {
			net::notice($sbnick, $rnick, 'Permission Denied');
			return;
		}

		if($msg =~ /^help/i) {
			my @args = split(/ /, $msg);
			shift @args; #discard first token 'help'
			help::sendhelp($sbnick, $rnick, lc $sbnick, @args);
		}

		elsif($msg =~ /^notice (\S*) (.*)/i) {
			net::notice($sbnick, $1, $2);
		}

		elsif($msg =~ /^msg (\S*) (.*)/i) {
			net::privmsg($sbnick, $1, $2);
		}

		elsif($msg =~ /^raw (.*)/i) {
			if(!adminserv::is_svsop($user, adminserv::S_ROOT() )) {
				notice($user, 'You do not have sufficient rank for this command');
				return;
			}
			net::ircsend($1);
		}

		elsif($msg =~ /^kill (\S*) (.*)/i) {
			net::irckill($sbnick, $1, $2);
		}

		elsif($msg =~ /^conf/i) {
			my @k = keys(%conf);
			my @v = values(%conf);

			net::notice($sbnick, $rnick, "Configuration:");
			for(my $i=0; $i<@k; $i++) {
				net::notice($sbnick, $rnick, $k[$i]." = ".$v[$i]);
			}
		}

		elsif($msg =~ /^set (\S+) (.*)/i) {
			if(!adminserv::is_svsop($user, adminserv::S_ROOT() )) {
				notice($user, 'You do not have sufficient rank for this command');
				return;
			}

			my @p = ($1, $2);
			chomp $p[1];

			if(update_conf($p[0], $p[1])) {
				net::notice($sbnick, $rnick, "Configuration: ".$p[0]." = ".$p[1]);
			} else {
				net::notice($sbnick, $rnick, "That value is read-only.");
			}
		}

		elsif($msg =~ /^save/i) {
			net::notice($sbnick, $rnick, "Saving configuration.");

			saveconf();
		}

		elsif($msg =~ /^rehash/i) {
			net::notice($sbnick, $rnick, "Loading configuration.");

			loadconf(1);
		}

		elsif($msg =~ /^tssync/i) {
			net::tssync();
		}

		elsif($msg =~ /^svsnick (\S+) (\S+)/i) {
			if(!adminserv::is_svsop($user, adminserv::S_ROOT() )) {
				notice($user, 'You do not have sufficient rank for this command');
				return;
			}
			net::svsnick($sbnick, $1, $2);
		}

		elsif($msg =~ /^tor-update/i) {
			net::notice($sbnick, $rnick, "Updating Tor server list.");
			update_tor_list();
		}
		elsif($msg =~ /^ss-update/i) {
			net::notice($sbnick, $rnick, "Updating SS definitions.");
			update_ss();
		}

		elsif($msg =~ /^tkl/i) {
			#$msg =~ s/^tkl list//i;
			#my @args = split(/ /, $msg);
			sb_tkl($user, $msg);
		}
}

sub noticehandle {
	my ($rnick, $dst, $msg) = @_;

	if($msg =~ /^\x01(\S+)\s?(.*?)\x01?$/) {
		net::privmsg($sbnick, $config::conf{diag}, "Got $1 reply from $rnick: $2");
	}
}

sub sb_tkl($$) {
# This function is a hack to fit better our normal services coding style.
# Better fix is to rewrite msghandle in another cleanup patch.
	my ($user, $msg) = @_;
	# We discard first token 'tkl'
	my (undef, $cmd, $msg) = split(/ /, $msg, 3);
	if(lc($cmd) eq 'list') {
		if($msg) {
			sb_tkl_glob($user, $msg);
		}
		else {
			sb_tkl_list($user);
		}
	}
	elsif(lc($cmd) eq 'del') {
		unless($msg) {
			nickserv::notice($user, "You have to specify at least one parameter");
		}
		sb_tkl_glob_delete($user, $msg);
	}
}

sub sb_tkl_list($) {
	my ($user) = @_;
	my @reply;
	$get_all_tklban->execute();
	while(my ($type, $ident, $host, $setter, $expire, $time, $reason) = $get_all_tklban->fetchrow_array()) {
		if($type eq 'Q') {
			#push @reply, "$type $host $setter";
			next;
		}
		else {
			push @reply, "$type $ident\@$host $setter";
		}
		$time = misc::gmtime2($time); $expire = misc::time_rel($expire - time()) if $expire;
		push @reply, "  set: $time; ".($expire ? "expires in: $expire" : "Will not expire");
		push @reply, "  reason: $reason";
	}
	$get_all_tklban->finish();
	push @reply, "No results" unless @reply;

	nickserv::notice($user, @reply);
}

sub sb_tkl_glob($$) {
	my ($user, $cmdline) = @_;

	my $sql_expr = "SELECT type, ident, host, setter, expire, time, reason FROM tklban WHERE ";

	my ($filters, $parms) = split(/ /, $cmdline, 2);
	my @filters = split(//, $filters);
	unless($filters[0] eq '+' or $filters[0] eq '-') {
		nickserv::notice($user, "Invalid Syntax: First parameter must be a set of filters preceded by a + or -");
		return;
	}
	my @args = misc::parse_quoted($parms);

	my ($success, $expr) = make_tkl_query(\@filters, \@args);
	unless ($success) {
		nickserv::notice($user, "Error: $expr");
		return;
	}
	$sql_expr .= "$expr ORDER BY type, time, host";

	my @reply;
	my $get_glob_tklban = $dbh->prepare($sql_expr);
	$get_glob_tklban->execute();
	while(my ($type, $ident, $host, $setter, $expire, $time, $reason) = $get_glob_tklban->fetchrow_array()) {
		if($type eq 'Q') {
			#push @reply, "$type $host $setter";
			next;
		}
		else {
			push @reply, "$type $ident\@$host $setter";
		}
		$time = misc::gmtime2($time); $expire = misc::time_rel($expire - time()) if $expire;
		push @reply, "  set: $time; ".($expire ? "expires in: $expire" : "Will not expire");
		push @reply, "  reason: $reason";
	}
	$get_glob_tklban->finish();

	push @reply, "No results" unless @reply;
	nickserv::notice($user, @reply);
}

sub sb_tkl_glob_delete($$) {
	my ($user, $cmdline) = @_;

	my $sql_expr = "SELECT type, ident, host FROM tklban WHERE ";

	my ($filters, $parms) = split(/ /, $cmdline, 2);
	my @filters = split(//, $filters);
	unless($filters[0] eq '+' or $filters[0] eq '-') {
		nickserv::notice($user, "Invalid Syntax: First parameter must be a set of filters preceded by a + or -");
		return;
	}
	my @args = misc::parse_quoted($parms);

	my ($success, $expr) = make_tkl_query(\@filters, \@args);
	unless ($success) {
		nickserv::notice($user, "Error: $expr");
		return;
	}

	$sql_expr .= $expr;

	my $src = $user->{ NICK }; # the nickserv function is kinda overkill here.
	my $get_glob_tklban = $dbh->prepare($sql_expr);
	$get_glob_tklban->execute();
	while(my ($type, $ident, $host) = $get_glob_tklban->fetchrow_array()) {
		if($type eq 'G') {
			net::unkline($src, $ident, $host);
		}
		elsif($type eq 'Z') {
			net::unzline($src, $host);
		}
	}
	$get_glob_tklban->finish();

}

sub make_tkl_query($$) {
	my ($parm1, $parm2) = @_;
	my @filters = @$parm1; my @args = @$parm2;

	my ($sign, $sql_expr);
	while(my $filter = shift @filters) {
		if ($filter eq '+') {
			$sign = +1;
			next;
		}
		elsif($filter eq '-') {
			$sign = 0;
			next;
		}

		my $parm = shift @args;
		unless (defined($parm)) {
			return (0, "Not enough arguments for filters.");
		}
		if($filter eq 'm') {
			my ($mnick, $mident, $mhost) = misc::parse_mask($parm);
			$mhost = $mnick if (!$mhost and !$mident);
			$mident = $dbh->quote($mident) if $mident;
			$mhost = $dbh->quote($mhost) if $mhost;
			$mhost =~ tr/*?/%_/; $mident =~ tr/*?/%_/;
			$sql_expr .= ($mident ? ($sign ? '' : '!').
				"(ident LIKE $mident) " : '').
				($mhost ? ($sign ? '' : '!').
				"(host LIKE $mhost) " : '');
		}
		elsif($filter eq 'r') {
			my $reason = $dbh->quote($parm);
			$reason =~ tr/\*\?/\%\_/;
			$sql_expr .= ($sign ? '' : '!')."(reason LIKE $reason) ";
			
		}
		elsif($filter eq 's') {
			my $setter = $dbh->quote($parm);
			$setter =~ tr/\*\?/\%\_/;
			$sql_expr .= ($sign ? '' : '!')."(setter LIKE $setter) ";
			
		}
		if($filter eq 'M') {
			my ($mnick, $mident, $mhost) = misc::parse_mask($parm);
			$mhost = $mnick if (!$mhost and !$mident);
			$mident = $dbh->quote($mident) if $mident;
			$mhost = $dbh->quote($mhost) if $mhost;
			$sql_expr .= ($mident ? ($sign ? '' : '!').
				"(ident REGEXP $mident) " : '').
				($mhost ? ($sign ? '' : '!').
				"(host REGEXP $mhost) " : '');
		}
		elsif($filter eq 'R') {
			my $reason = $dbh->quote($parm);
			$sql_expr .= ($sign ? '' : '!')."(reason REGEXP $reason) ";
			
		}
		elsif($filter eq 'S') {
			my $setter = $dbh->quote($parm);
			$sql_expr .= ($sign ? '' : '!')."(setter REGEXP $setter) ";
			
		}
		$sql_expr .= 'AND ' if (@filters);
	}
	if (scalar(@args)) {
		return (0, "Too many arguments for filters.");
	}
	return (1, $sql_expr);
}

sub get_tkl_type_name($) {
	my %tkltype = (
		G => 'G:line',
		Z => 'GZ:line',
		s => 'Shun',
		Q => 'Q:line',
	);
	return $tkltype{$_[0]};
};

sub get_filter_action_name($) {
	my %filteraction = (
		Z => 'GZ:line',
		S => 'tempshun',
		s => 'shun',
		g => 'G:line',
		z => 'Z:line',
		k => 'K:line',
		K => 'Kill',
		b => 'Block',
		d => 'DCC Block',
		v => 'Virus Chan',
		w => 'Warn',
		#t => 'Test', # Should never show up, and not implemented in 3.2.4 yet.
	);
	return $filteraction{$_[0]};
};

sub handle_tkl($$@) {
	my ($type, $sign, @parms) = @_;
	if ($type eq 'G' or $type eq 'Z' or $type eq 's' or $type eq 'Q') {
		if ($sign == +1) {
			my ($ident, $host, $setter, $expire, $time, $reason) = @parms;
			$add_tklban->execute($type, $ident, $host, $setter, $expire, $time, $reason);
			$add_tklban->finish();
			net::privmsg($sbnick, $config::conf{diag}, get_tkl_type_name($type)." added for $ident\@$host ".
				"from ($setter on ".misc::gmtime2($time).' to expire at '.misc::gmtime2($expire).": $reason)")
					if net::synced() and $type ne 'Q';
		}
		elsif($sign == -1) {
			my ($ident, $host, $setter) = @parms;

			if ($type ne 'Q' and net::synced()) {
				$get_tklban->execute($type, $ident, $host);
				my (undef, $expire, $time, $reason) = $get_tklban->fetchrow_array;
				$get_tklban->finish();

				net::privmsg($sbnick, $config::conf{diag}, "$setter removed ".get_tkl_type_name($type)." $ident\@$host ".
					"set at ".misc::gmtime2($time)." - reason: $reason");
			}

			$del_tklban->execute($type, $ident, $host);
			$del_tklban->finish();
		}
	}
	elsif($type eq 'F') {
		if($sign == +1) {
			my ($target, $action, $setter, $expire, $time, $bantime, $reason, $mask) = @parms;
			$add_spamfilter->execute($target, $action, $setter, $expire, $time, $bantime, $reason, $mask);
			$add_spamfilter->finish();
			net::privmsg($sbnick, $config::conf{diag}, "Spamfilter added: '$mask' [target: $target] [action: ".
				get_filter_action_name($action)."] [reason: $reason] on ".misc::gmtime2($time)."from ($setter)")
					if net::synced();
		}
		elsif($sign == -1) {
			# TKL - F u Z tabris!northman@tabris.netadmin.SCnet.ops 0 0 :do_not!use@mask
			my ($target, $action, $setter, $mask) = @parms;
			if(net::synced()) {
				$get_spamfilter->execute($target, $action, $mask);
				my ($time, $reason) = $get_spamfilter->fetchrow_array;
				$get_spamfilter->finish();
				$reason =~ tr/_/ /;
				net::privmsg($sbnick, $config::conf{diag}, "$setter removed Spamfilter (action: ".get_filter_action_name($action).
					", targets: $target) (reason: $reason) '$mask' set at: ".misc::gmtime2($time));
			}
			$del_spamfilter->execute($target, $action, $mask);
			$del_spamfilter->finish();
		}
	}
}

sub saveconf() {
	config::writeHash(\%conf, "config/securitybot/sb.conf");
}

sub loadconf($) {
	my ($update) = @_;
	
	%conf = config::readHash("config/securitybot/sb.conf");
	%sql_conf = config::readHash("config/securitybot/sb.mysql");
}

sub update_conf($$) {
	my ($k, $v) = @_;

	return 0 if($k eq 'EnableTor' or $k eq 'EnableCountry');

	$conf{$k} = $v;
	return 1;
}

########################################################################
#                       SecureServ definitions                         #
########################################################################
my $ss_base_url = 'http://secure.irc-chat.net/defs.php';

sub get_ss_definitions() {
	my $url = "'$ss_base_url?u=".$conf{'ss_Username'}.'&p='.$conf{'ss_Pass'}."'";

	my %skipfilters;
	foreach my $skipfilter (split(/ /, $conf{'ss_SkipFilters'})) {
		$skipfilters{uc($skipfilter)} = 1;
	}
	
	open DATAFILE, '-|', ('wget -q -O - ' . $url);
	#open DATAFILE, "$main::prefix/data/viri.dat";

	my (@filters, $version);
	while (my $l = <DATAFILE>) {
	#XcoderBot 2 0 0 "^Xcoder.*" "You appear to be infected with a BotNet Trojan" 1
		chomp $l;
		if($l =~ /^(\S+) (\d) (\d) (\d) \"(.+)\" \"(.+)\" (\d)$/) {
			my ($filter, $type, $regex, $reason, $action) = ($1,$2,$5,$6,$7);
			next if $skipfilters{uc($filter)};
			my $ref = make_ss_tkl($filter, $type, $regex, $reason, $action);
			push @filters, $ref if $ref;
		}
		if ($l =~ /^(\d+)$/) {
			$version = $1;
		}
	}
	return ($version, @filters);
}

sub make_ss_tkl($$$$$) {
	my ($filter, $type, $regex, $reason, $action) = @_;
	my ($tkl_target, $tkl_action);
	$regex =~ s/^\s+//;
	$reason =~ s/ /_/g;
	if($regex =~ /\(\?\i\)/) {
		$regex =~ s/\(\?\i\)//;
	}
	else {
		$regex = "(?i)$regex";
	}
	if($type == 0) {
		if ($regex =~ /^\^/ or $regex =~ /\$$/) {
			$regex =~ s/^\^/\^\x01/;
			$regex =~ s/\$$/\x01\$/;
		}
		else {
			$regex = "\x01$regex\x01";
		}
		$tkl_target = 'n';
	}
	elsif($type == 1 or $type == 6) {
		$tkl_target = 'cpnNPq';
	}
	elsif($type == 2) {
		$tkl_target = 'u';
		$regex =~ s/\$$//;
		$regex = "$regex!";
	}
	elsif($type == 3) {
		$tkl_target = 'u';
		$regex =~ s/\$$//;
		$regex =~ s/^\^//;
		$regex = "!$regex@";
	}
	elsif($type == 4) {
		$tkl_target = 'u';
		$regex =~ s/^\^//;
		$regex = ":$regex";
	}
	elsif($type == 5) {
		# We don't have a spamfilter type for this
		# this would be a drone channel.
		return undef;
	}

	if($action == 0) {
		$tkl_action = 'v'; # viruschan
	}
	elsif($action == 1) {
		$tkl_action = 'Z'; # gzline. gline would be better emulation
	}
	elsif($action == 2 or $action == 3) {
		$tkl_action = 'b'; # block.
		# would be better to trap but not block, but there is no such spamfilter action.
	}

	return [$tkl_target, $tkl_action, "$filter:_$reason", $regex];
}

########################################################################
#   END                 SecureServ definitions                   END   #
########################################################################

sub update_ss_timed($) {
	my ($time) = @_;

	event::addtimer('3600', $time, __PACKAGE__, "securitybot::update_ss_timed");

	update_ss();
}

my $last_version : shared;
sub update_ss() {
	my %ss_filters;
	net::privmsg($sbnick, $config::conf{diag}, " -- Loading SecureServ definitions.");
	$get_all_spamfilter->execute();
	while(my ($target, $action, $setter, $expire, $time, $bantime, $reason, $regex) =
		$get_all_spamfilter->fetchrow_array())
	{
		next unless $setter eq $sbnick;
		$ss_filters{"$target $action $reason $regex"} = [$target, $action, $reason, $regex];
	}
	$get_all_spamfilter->finish();

	my ($version, @filters) = get_ss_definitions();
	if ($last_version and $version <= $last_version) {
		net::privmsg($sbnick, $config::conf{diag}, " -- SecureServ definitions already up-to-date.");
		return;
	}
	$last_version = $version;
	unless (@filters) {
		net::privmsg($sbnick, $config::conf{diag}, " -- Loading SecureServ definitions failed.");
		return;
	}

	foreach my $tmp (@filters) {
		my ($target, $action, $reason, $regex) = @$tmp;
		if(!defined($ss_filters{"$target $action $reason $regex"})) {
			net::spamfilter(+1, $target, $action, $sbnick, $conf{'DroneKlineTime'}, $reason, $regex);
		}
		delete($ss_filters{"$target $action $reason $regex"});
	}

	foreach my $filter_str (keys(%ss_filters)) {
		my ($target, $action, $reason, $regex) = @{$ss_filters{$filter_str}};
		net::spamfilter(0, $target, $action, $sbnick, $conf{'DroneKlineTime'}, $reason, $regex);
	}
	net::privmsg($sbnick, $config::conf{diag}, " -- Loading SecureServ definitions succeeded.");
	
}

sub expire_tkl() {
	$get_expired_tklban->execute();
	while (my ($type, $ident, $host, $setter, $expire, $time, $reason) = $get_expired_tklban->fetchrow_array()) {
		if ($type eq 'G' or $type eq 'Z' or $type eq 's') {
			net::privmsg($sbnick, $config::conf{diag}, "Expiring ".get_tkl_type_name($type)." $ident\@$host ".
				"set by $setter at ".misc::gmtime2($time)." - reason: $reason");
			#$del_tklban->execute($type, $ident, $host);
			#$del_tklban->finish();
			}
	}
	$get_expired_tklban->finish();

	$del_expired_tklban->execute();
	$del_expired_tklban->finish();
}

sub expire_tkl_timed {
	my ($time) = @_;
	$time = 10 unless $time;

	event::addtimer('10', $time, __PACKAGE__, "securitybot::expire_tkl_timed");

	expire_tkl();
}

sub end { }
sub unload { saveconf(); }

1;
